#ifndef BINALLOCATOR_H_VOR042
#define BINALLOCATOR_H_VOR042

#include "TestAlgor.h"
#include "Particle.h"
#include <boost/timer.hpp>
#include <list>
#include <iostream>
#include <limits>

template <typename T>
class BinAllocatorImpl {
public:
  typedef size_t size_type;
  typedef ptrdiff_t difference_type;
  typedef T* pointer;
  typedef const T* const_pointer;
  typedef T& reference;
  typedef const T& const_reference;
  typedef T value_type;

  BinAllocatorImpl() : cached(0), refCount( 1 ) {
    //cout << typeid( *this ).name() << " constructed." << endl;
    baskets.push_back( Basket() );
    lastBasket = baskets.begin();
    lastBin = lastBasket->begin();
  }

  ~BinAllocatorImpl() {
    //cout << typeid( *this ).name() << " destructed." << endl;
  }

  pointer allocate() {
    //early exit if we have a cached
    if ( cached ) {
      Bin* temp = cached;
      cached = 0;
      return temp->markAndGet();
    }

    //early exit if we are on an empty bin
    if ( lastBin->isFree() )
      return lastBin->markAndGet();

    BinIter binStart = lastBin;

    do {
      ++lastBin;

      //implement wrap-arounds
      if ( lastBin == lastBasket->end() ) {
        //Move on to next basket
        ++lastBasket;
        //Even if there is one basket, this is OK, we go to start.
        if ( lastBasket == baskets.end() )
          lastBasket = baskets.begin();
        lastBin = lastBasket->begin();
      }

      if ( lastBin->isFree() )
        return lastBin->markAndGet();
    } while ( lastBin != binStart );

    //If we get here, that means there were no empty bins
    baskets.push_back( Basket() );
    lastBasket = --baskets.end();
    lastBin = lastBasket->begin() + 1; //set to second element

    return lastBasket->begin()->markAndGet();
  }

  void deallocate( pointer p ) {
    //cout << p << " deallocated." << endl;
    Bin* bin = (Bin*)((char*)p-4);
    bin->unmark();
    cached = bin;
  }

  void Acquire() {
    ++refCount;
    //cout << typeid( *this ).name() << " acquired, count = " << refCount << endl;
  }

  void Release() {
    //cout << typeid( *this ).name() << " released, count = " << refCount-1 << endl;
    if ( --refCount <= 0 )
      delete this;
  }

private:
  struct Bin {
    char memory[ sizeof(T)+4 ];

    Bin() { memory[0] = 0; }

    bool isFree() {
      return memory[0] == 0;
    }

    pointer markAndGet() {
      assert( isFree() );
      memory[0] = 1;
      return reinterpret_cast<pointer>( &memory[4] );
    }

    void unmark() {
      memory[0] = 0;
    }
  };

  //This size must be at least 2!
  static const int BASKET_SIZE = 500;

  class Basket {
    Bin bins[ BASKET_SIZE ];

  public:

    typedef Bin* iterator;

    iterator begin() {
      return bins;
    }

    iterator end() {
      return bins + BASKET_SIZE;
    }
  };

  typedef std::list<Basket> BasketList;
  typedef typename BasketList::iterator BasketListIter;
  typedef typename Basket::iterator BinIter;

  BasketList baskets;
  BasketListIter lastBasket;
  BinIter lastBin;

  Bin* cached;

  int refCount;
};

template <typename T> class BinAllocator;

template <> class BinAllocator<void> {
public:
  typedef void* pointer;
  typedef const void* const_pointer;
  // reference to void members are impossible.
  typedef void value_type;

  template <class U> struct rebind { typedef BinAllocator<U> other; };
};

template <typename T>
class BinAllocator {
public:
  typedef size_t size_type;
  typedef ptrdiff_t difference_type;
  typedef T* pointer;
  typedef const T* const_pointer;
  typedef T& reference;
  typedef const T& const_reference;
  typedef T value_type;
  template <typename U> struct rebind { typedef BinAllocator<U> other; };

  BinAllocator() throw() : p_impl(0) {
    //cout << typeid( *this ).name() << " constructed." << endl;
  }

  BinAllocator( const BinAllocator& o ) throw() : p_impl( o.p_impl ) {
    //cout << typeid( *this ).name() << " copied constructed." << endl;
    if ( p_impl )
      p_impl->Acquire();
  }

  template <typename U>
  BinAllocator( const BinAllocator<U>& ) throw() : p_impl(0) {
    //cout << typeid( *this ).name() << " constructed." << endl;
    //OK, we have to do a conversion.  Ignore our parameter.
  }

  ~BinAllocator() throw() {
    //cout << typeid( *this ).name() << " destroyed." << endl;
    if ( p_impl )
      p_impl->Release();
  }

  pointer address( reference x ) const {
    return &x;
  }
  
  const_pointer address( const_reference x ) const {
    return &x;
  }

  pointer allocate( size_type num, BinAllocator<void>::const_pointer hint = 0 ) {
    if ( !p_impl )
      p_impl = new BinAllocatorImpl<T>();

    assert( num == 1 );

    pointer ret = p_impl->allocate();
    //cout << ret << " allocated." << endl;
    return ret;
  }

  void deallocate( pointer p, size_type num ) {
    assert( p_impl );
    p_impl->deallocate( p );
    assert( num == 1 );
  }

  size_type max_size() const throw() {
    return std::numeric_limits<size_type>::max();
  }

  void construct( pointer p, const T& val ) {
    new((void*)p) T(val);
  }

  void destroy( pointer p ) {
    ((T*)p)->~T();
  }

  BinAllocator& operator = ( const BinAllocator& rhs ) {
    //cout << typeid( *this ).name() << " copied." << endl;
    if ( p_impl )
      p_impl->Release();

    p_impl = rhs.p_impl;

    if ( p_impl )
      p_impl->Acquire();
  }

private:
  BinAllocatorImpl<T>* p_impl;
};

#endif //BINALLOCATOR_H_VOR042
